class PyPokedexError(RuntimeError):
    pass


class PyPokedexHTTPError(PyPokedexError):
    def __init__(self, message: str, http_code: int) -> None:
        super().__init__(message)
        self._http_code = http_code

    @property
    def http_code(self) -> int:
        return self._http_code



from collections import defaultdict
from typing import DefaultDict, List, NamedTuple, Dict, Optional

from pypokedex.exceptions import PyPokedexError


class Ability(NamedTuple):
    name: str
    is_hidden: bool


class BaseStats(NamedTuple):
    hp: int
    attack: int
    defense: int
    sp_atk: int
    sp_def: int
    speed: int


class Move(NamedTuple):
    name: str
    learn_method: str
    level: Optional[int]


class Sprites(NamedTuple):
    front: Dict[str, Optional[str]]
    back: Dict[str, Optional[str]]


class Pokemon:
    dex: int
    name: str
    height: int
    weight: int
    base_experience: int
    base_stats: BaseStats
    abilities: List[Ability]
    types: List[str]
    moves: DefaultDict[str, List[Move]]
    sprites: Sprites

    def __init__(self, json_data) -> None:
        """Loads and stores required pokemon data"""

        # pylint: disable=too-many-locals
        try:
            self.dex = json_data["id"]

            for pokemon_info in ["name", "height", "weight", "base_experience"]:
                setattr(self, pokemon_info, json_data[pokemon_info])

            stat_dict = {}

            for stat in json_data["stats"]:
                stat_value = stat["base_stat"]
                stat_name = stat["stat"]["name"]

                if stat_name in ["hp", "attack", "defense", "speed"]:
                    stat_dict[stat_name] = stat_value
                elif stat_name == "special-attack":
                    stat_dict["sp_atk"] = stat_value
                elif stat_name == "special-defense":  # pragma: no branch
                    stat_dict["sp_def"] = stat_value

            self.base_stats = BaseStats(**stat_dict)

            self.abilities = [
                Ability(ability["ability"]["name"], ability["is_hidden"])
                for ability in json_data["abilities"]
            ]

            self.types = [type_["type"]["name"] for type_ in json_data["types"]]

            self.moves = defaultdict(list)

            for move in json_data["moves"]:
                move_name = move["move"]["name"]

                for game_details in move["version_group_details"]:
                    learn_level = game_details["level_learned_at"]
                    learn_method = game_details["move_learn_method"]["name"]
                    game_name = game_details["version_group"]["name"]
                    if learn_level == 0:  # Move not learned by level-up
                        learn_level = None

                    self.moves[game_name].append(
                        Move(move_name, learn_method, learn_level)
                    )

            self.sprites = Sprites(front={}, back={})
            for sprite, url in json_data["sprites"].items():
                sprite_direction, sprite_type = sprite.split("_", 1)

                if sprite_direction == "front":
                    self.sprites.front[sprite_type] = url
                else:
                    self.sprites.back[sprite_type] = url

        except KeyError as error:
            raise PyPokedexError(
                "A required piece of data was not found for the current Pokemon!"
            ) from error

    def exists_in(self, game: str) -> bool:
        """Checks whether the current Pokemon exists in the specified game."""
        return game in self.moves

    def learns(self, move_name: str, game: str) -> bool:
        """Checks whether the current Pokemon learn the specified move
        in the specified game."""
        if not self.exists_in(game):
            raise PyPokedexError(f"{self.name} is not obtainable in {game}!")

        for move in self.moves[game]:
            if move.name == move_name:
                return True

        return False

    def __str__(self) -> str:
        """Returns a human-readable representation of the current Pokemon."""
        return f"Pokemon(dex={self.dex}, name='{self.name}')"

    def __eq__(self, other) -> bool:
        return self.dex == other.dex

    def __lt__(self, other) -> bool:
        return self.dex < other.dex

    def __gt__(self, other) -> bool:
        return self.dex > other.dex

    def __le__(self, other) -> bool:
        return self.dex <= other.dex

    def __ge__(self, other) -> bool:
        return self.dex >= other.dex



from functools import lru_cache
from typing import Union

import requests

from pypokedex.exceptions import PyPokedexError, PyPokedexHTTPError
from pypokedex.pokemon import Pokemon

POKEAPI_BASE_URL = "https://pokeapi.co/api/v2/pokemon"


@lru_cache(maxsize=None)
def get(**kwargs) -> Pokemon:
    if len(kwargs) != 1:
        raise TypeError("pypokedex.get() expects expects only 1 argument!")

    subpage: Union[int, str]

    if "dex" in kwargs and isinstance(kwargs["dex"], int):
        subpage = kwargs["dex"]
    elif "name" in kwargs and isinstance(kwargs["name"], str):
        subpage = kwargs["name"].lower()
    else:
        raise TypeError("Arguments were either of an incorrect type or value!")

    try:
        response = requests.get(f"{POKEAPI_BASE_URL}/{subpage}", timeout=3)
        response.raise_for_status()

    except requests.exceptions.HTTPError as error:
        if response.status_code == 404:
            raise PyPokedexHTTPError(
                f"The requested pokemon was not found!", 404
            ) from error
        raise PyPokedexHTTPError(
            f"An HTTP error occurred! (Status code: {response.status_code})",
            response.status_code,
        ) from error

    except requests.exceptions.RequestException as error:
        raise PyPokedexError("An internal requests exception occurred!") from error

    return Pokemon(response.json())



import os
import re

from setuptools import setup, find_packages

this_directory = os.path.abspath(os.path.dirname(__file__))

# Get long description from README.md
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as readme_file:
    long_description = readme_file.read()

# Get package version from pypokedex/__init__.py
with open(os.path.join(this_directory, "pypokedex", "__init__.py")) as version_file:
    version = re.search(
        r'^__version__ = [\'"]([^\'"]*)[\'"]', version_file.read(), re.M
    ).group(1)

setup(
    name="pypokedex",
    version=version,
    description="A minimal pokedex library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/arnavb/pypokedex",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Games/Entertainment",
        "Topic :: Software Development :: Libraries",
    ],
    author="Arnav Borborah",
    author_email="arnavborborah11@gmail.com",
    license="MIT",
    packages=find_packages(exclude=["tests", "docs"]),
    install_requires=["requests>=2.21.0"],
    python_requires=">=3.6",
)
